import base64

try:
    # noinspection PyPep8Naming
    import cPickle as pickle
except ImportError:
    import pickle
import types
import typing as tp
import zlib

from satella.json import JSONAble


class GenerationPolicy:
    """
    A policy that manages generating a traceback

    For a "default" policy it's pretty customizable

    Override if need be, and pass the class (or instance) to Traceback

    :param enable_pickling: bool, whether to enable pickling at all
    :param compress_at: pickles longer than this (bytes) will be compressed
    :param repr_length_limit: maximum p_len of __repr__. None for no limit.
    :param compression_level: "1" is fastest, "9" is slowest
    """
    __slots__ = ('enable_pickling', 'compress_at', 'repr_length_limit', 'compression_level')

    def __init__(self, enable_pickling: bool = True,
                 compress_at: int = 128 * 1024,
                 repr_length_limit: int = 128 * 1024,
                 compression_level: int = 6):
        assert isinstance(compression_level, int) and 1 <= compression_level <= 9

        self.enable_pickling = enable_pickling  # type: bool
        self.compress_at = compress_at  # type: int
        self.repr_length_limit = repr_length_limit  # type: int
        self.compression_level = compression_level  # type: int

    def should_pickle(self, value: tp.Any) -> bool:
        """
        Should this value be pickled?

        :param value: value candidate
        :return: bool
        """
        return self.enable_pickling

    def should_compress(self, pickle_data: bytes) -> bool:
        """
        Should this pickle undergo compression?

        :param pickle_data: pickle value
        :return: bool
        """
        return len(pickle_data) > self.compress_at

    def get_compression_level(self, pickledata: bytes) -> int:
        """
        What compression level to use to pickle this?

        :param pickledata: bytes, pickle value
        :return: int, 1-9, where "1" is the fastest, and "9" is the slowest,
            but produces best compression
        """
        return self.compression_level

    def process_repr(self, r: str) -> str:
        """
        Process the string obtained from __repr__ing

        :param r: result of a __repr__ on value
        :return: processed result
        """
        if self.repr_length_limit is None:
            return r
        else:
            if len(r) > self.repr_length_limit:
                return r[:self.repr_length_limit] + u'...'
            else:
                return r


class StoredVariableValue(JSONAble):
    """
    Class used to store a variable value. Picklable.

    Attributes are:
        .repr - a text representation obtained using repr
        .typeinfo - a text representation of variable's type
        .pickle - bytes with pickled (optionally processed) value, or None if
            not available
        .pickle_type - what is stored in .pickle?
            None - nothing
            "pickle" - normal Python pickle
            "pickle/gzip" - Python pickle treated with zlib.compress
            "failed" - could not pickle, pickle contains a UTF-8 text with
                human-readable exception reason
            "failed/gzip" - compression failed, pickle contains a UTF-8 text with
                human-readable exception reason

    If value cannot be pickled, it's repr will be at least preserved.

    Note that the value itself won't be preserved.

    :param value: any Python value to preserve
    :param policy: policy to use (instance)
    """
    __slots__ = ('repr', 'type_', 'pickle', 'pickle_type')

    @classmethod
    def from_json(cls, x: dict) -> 'StoredVariableValue':
        sv = StoredVariableValue.__new__(StoredVariableValue)
        sv.repr = x['repr']
        sv.type_ = x['type']
        sv.pickle_type = x['pickle_type']
        if 'pickle' in x:
            sv.pickle = base64.b64decode(x['pickle'].encode('utf8'))
        sv.pickle = None
        return sv

    def to_json(self) -> dict:
        dct = {
            'repr': self.repr,
            'type': self.type_,
            'pickle_type': self.pickle_type
        }
        if self.pickle:
            dct['pickle'] = base64.b64encode(self.pickle).decode('utf8')
        return dct

    def __init__(self, value: tp.Any, policy: tp.Optional[GenerationPolicy] = None):
        self.repr = repr(value)  # type: str
        self.type_ = repr(type(value))  # type: str

        policy = policy or GenerationPolicy()

        self.repr = policy.process_repr(self.repr)

        self.pickle = None  # type: bytes
        self.pickle_type = None  # type: str

        if policy.should_pickle(value):
            try:
                self.pickle = pickle.dumps(value, pickle.HIGHEST_PROTOCOL)
                self.pickle_type = 'pickle'
            except BaseException as e:
                # yes, they all can happen!
                self.pickle = repr((e,) + e.args).encode('utf8')
                self.pickle_type = "failed"
            else:
                if policy.should_compress(self.pickle):
                    try:
                        self.pickle = zlib.compress(
                            self.pickle,
                            policy.get_compression_level(
                                self.pickle))
                        self.pickle_type = "pickle/gzip"
                    except zlib.error as e:
                        self.pickle = ('failed to gzip, reason is %s' % (repr(e),)).encode('utf8')
                        self.pickle_type = "failed/gzip"

    def load_value(self):
        """
        Return the value that this represents.

        WARNING! This may result in importing things from environment, as
        pickle.loads will be called.

        :return: stored value - if picklable and was pickled
        :raises ValueError: value has failed to be pickled or was never pickled
        """
        if self.pickle_type is None:
            raise ValueError('value was never pickled')
        elif self.pickle_type == 'failed':
            raise ValueError(
                'MemoryCondition has failed to be pickled, reason is %s' % (self.pickle,))
        elif self.pickle_type == 'pickle/gzip':
            pickle_ = zlib.decompress(self.pickle)
        elif self.pickle_type == 'pickle':
            pickle_ = self.pickle
        else:
            raise ValueError('unknown pickle type of %s' % (self.pickle_type,))

        try:
            return pickle.loads(pickle_)
        except pickle.UnpicklingError:
            raise ValueError(
                'object picklable, but cannot load in this environment')


class StackFrame(JSONAble):
    """
    Class used to verily preserve stack frames. Picklable.
    """
    __slots__ = ('locals', 'globals', 'name', 'filename', 'lineno')

    def __init__(self, frame: types.FrameType, policy: GenerationPolicy):
        self.name = frame.f_code.co_name  # type: str
        self.filename = frame.f_code.co_filename  # type: str
        self.lineno = frame.f_lineno  # type: int

        self.locals = {}  # type: tp.Dict[str, StoredVariableValue]
        for key, value in frame.f_locals.items():
            self.locals[key] = StoredVariableValue(value, policy)

        self.globals = {}  # type: tp.Dict[str, StoredVariableValue]
        for key, value in frame.f_globals.items():
            self.globals[key] = StoredVariableValue(value, policy)

    @classmethod
    def from_json(cls, x: dict) -> 'StackFrame':
        sv = StackFrame.__new__(StackFrame)
        sv.name = x['name']
        sv.filename = x['filename']
        sv.lineno = x['line_no'],
        sv.locals = {k: StoredVariableValue.from_json(v) for k, v in x['locals'].items()}
        sv.globals = {k: StoredVariableValue.from_json(v) for k, v in x['globals'].items()}
        return sv

    def to_json(self) -> dict:
        return {
            'name': self.name,
            'filename': self.filename,
            'line_no': self.lineno,
            'locals': {k: v.to_json() for k, v in self.locals.items()},
            'globals': {k: v.to_json() for k, v in self.globals.items()}
        }



import collections


class RollingArithmeticAverage:
    """
    A class to implement a rolling arithmetic average over n last entries

    :param n: amount of last entries to count
    """
    __slots__ = ('n', 'queue', 'tot_sum')

    def __init__(self, n: int = 100):
        self.n = n
        self.queue = collections.deque()
        self.tot_sum = 0

    def insert(self, x: float) -> None:
        """
        Add a value to the rolling average, discarding the previous entry if the
        buffer size is exceeded

        :param x: sample to insert
        """
        if len(self.queue) >= self.n:
            prev_entry = self.queue.popleft()
            self.tot_sum -= prev_entry

        self.queue.append(x)
        self.tot_sum += x

    def avg(self) -> float:
        """
        Compute current average

        :return: current average
        :raises ZeroDivisionError: the average buffer is empty
        """
        return self.tot_sum / len(self.queue)

    def clear(self) -> None:
        """
        Clear the rolling average buffer
        """
        self.queue.clear()
        self.tot_sum = 0





import logging
from concurrent.futures import Future as PythonFuture
from concurrent.futures._base import CANCELLED, CANCELLED_AND_NOTIFIED, FINISHED
import concurrent.futures._base

try:
    from concurrent.futures import InvalidStateError
except ImportError:
    class InvalidStateError(Exception):
        pass

LOGGER = logging.getLogger('concurrent.futures')
PRE_FINISHED = 'PRE-FINISHED'

concurrent.futures._base._STATE_TO_DESCRIPTION_MAP[PRE_FINISHED] = 'pre-finished'
concurrent.futures._base._FUTURE_STATES.append(PRE_FINISHED)


class Future(PythonFuture):
    """
    A future that allows it's callback handlers to change it's result before presenting
    it to the user.

    Use like this:

    >>> fut = Future()
    >>> fut.set_running_or_notify_cancel()
    >>> def transform_future(future):
    >>>     future.set_result(future.result() + 2)
    >>> fut.add_pre_done_callback(transform_future)
    >>> fut.set_result(2)
    >>> assert fut.result() == 4
    """
    def __init__(self):
        super().__init__()
        self._pre_done_callbacks = []

    def add_pre_done_callback(self, fn):
        """
        Attaches a callable that will be called just before the future finishes
        and can change the future's result (or insert an Exception).

        Args:
            fn: A callable that will be called with this future as its only
                argument just before the future completes or is cancelled.
        """
        with self._condition:
            if self._state not in [CANCELLED, CANCELLED_AND_NOTIFIED, FINISHED]:
                self._pre_done_callbacks.append(fn)
                return
        try:
            fn(self)
        except Exception:
            LOGGER.exception('exception calling callback for %r', self)

    def _invoke_pre_callbacks(self):
        for callback in self._pre_done_callbacks:
            try:
                callback(self)
            except Exception:
                LOGGER.exception('exception calling callback for %r', self)

    def set_result(self, result) -> None:
        if self._state == PRE_FINISHED:
            self._result = result
            self._exception = None
        else:
            super().set_result(result)

    def set_exception(self, exception) -> None:
        if self._state == PRE_FINISHED:
            self._exception = exception
            self._result = None
        else:
            super().set_exception(exception)

    def result(self, timeout=None):
        if self._state == PRE_FINISHED:
            return self.__get_result()
        else:
            return super().result(timeout)

    def exception(self, timeout: None):
        if self._state == PRE_FINISHED:
            return self._exception
        else:
            return super().exception(timeout)

    def set_result(self, result):
        """Sets the return value of work associated with the future.

        Should only be used by Executor implementations and unit tests.
        """
        if self._state == PRE_FINISHED:
            self._result = result
        else:
            with self._condition:
                if self._state in {CANCELLED, CANCELLED_AND_NOTIFIED, FINISHED}:
                    raise InvalidStateError('{}: {!r}'.format(self._state, self))
                self._result = result
                self._state = PRE_FINISHED
                self._invoke_pre_callbacks()
                self._state = FINISHED
                for waiter in self._waiters:
                    waiter.add_result(self)
                self._condition.notify_all()
            self._invoke_callbacks()

    def set_exception(self, exception):
        """Sets the result of the future as being the given exception.

        Should only be used by Executor implementations and unit tests.
        """
        if self._state == PRE_FINISHED:
            self._exception = exception
            self._result = None
        else:
            with self._condition:
                if self._state in {CANCELLED, CANCELLED_AND_NOTIFIED, FINISHED}:
                    raise InvalidStateError('{}: {!r}'.format(self._state, self))
                self._exception = exception
                self._state = PRE_FINISHED
                self._invoke_pre_callbacks()
                self._state = FINISHED
                for waiter in self._waiters:
                    waiter.add_exception(self)
                self._condition.notify_all()
            self._invoke_callbacks()


class WrappingFuture(Future):
    """
    A Satella future wrapping an existing Python future.

    Use like:

    >> wrapped = WrappingFuture(existing_python_future)
    """
    def __init__(self, source_future: PythonFuture):
        super().__init__()
        self.source_future = source_future
        self.source_future.add_done_callback(self._on_future_completed)
        self._old_state = source_future._state

    @property
    def _state(self):
        if self.source_future._state == FINISHED:
            return self._old_state
        else:
            return self.source_future._state

    @_state.setter
    def _state(self, v: str):
        self._old_state = v

    def set_running_or_notify_cancel(self) -> bool:
        return self.source_future.set_running_or_notify_cancel()

    def _on_future_completed(self, future: PythonFuture):
        if future._exception is not None:
            self.set_exception(future._exception)
        else:
            self.set_result(future._result)

    def cancel(self) -> bool:
        super().cancel()
        return self.source_future.cancel()





import typing as tp
import operator

__all__ = ['x']

import warnings

PredicateType = tp.Callable[[tp.Any], bool]


def _nop(v: tp.Any) -> tp.Any:
    return v


def make_operation_two_args(operation_two_args: tp.Callable[[tp.Any, tp.Any], tp.Any],
                            docstring: tp.Optional[str] = None,
                            swap_order: bool = False) -> PredicateType:
    def operation(self, a) -> PredicateType:
        if isinstance(a, Predicate):
            if swap_order:
                def op(v):
                    return operation_two_args(a(v), self(v))
                return Predicate(op)
            else:
                def op(v):
                    return operation_two_args(self(v), a(v))
                return Predicate(op)
        else:
            if swap_order:
                def op(v):
                    return operation_two_args(a, self(v))
                return Predicate(op)
            else:
                def op(v):
                    return operation_two_args(self(v), a)
                return Predicate(op)

    if docstring:
        operation.__doc__ = docstring

    return operation


def make_operation_single_arg(operation: tp.Callable[[tp.Any], tp.Any],
                              docstring: tp.Optional[str] = None) -> PredicateType:
    def operation_v(self) -> PredicateType:
        def operate(v):
            return operation(self.operation(v))
        return Predicate(operate)

    if docstring:
        operation_v.__doc__ = docstring

    return operation_v


def _has_keys(a: tp.Container, keys) -> bool:
    for key in keys:
        if key not in a:
            return False
    return True


def _one_of(a, values: tp.Container) -> bool:
    return a in values


class Predicate:
    """
    A shorthand to create lambdas using such statements, for example:

    >>> add_two = x + 2
    >>> assert add_two(2) == 4
    """
    __slots__ = ('operation', )

    def __init__(self, operation: tp.Callable[[tp.Any], tp.Any] = _nop):
        self.operation = operation

    str = make_operation_single_arg(str, 'Call str() on predicate')
    int = make_operation_single_arg(int, 'Call int() on predicate')
    float = make_operation_single_arg(float, 'Call float() on predicate')

    def __call__(self, *args) -> bool:
        if len(args) == 0:
            return Predicate(lambda y: self.operation(y)())
        else:
            return self.operation(args[0])

    def has_keys(self, *keys) -> PredicateType:
        """
        Return a predicate checking whether this value has provided keys
        """
        return make_operation_two_args(_has_keys)(self, keys)

    def one_of(self, *values) -> PredicateType:
        """
        Return a predicate checking if x is amongst values
        """
        return make_operation_two_args(_one_of)(self, values)

    def has_p(self, predicate: 'Predicate') -> PredicateType:
        """
        An old name for has().

        It's deprecated. Use has() instead
        """
        warnings.warn('This is deprecated and will be removed in Satella 3.x.'
                      'Please use has() instead', DeprecationWarning)
        return self.has(predicate)

    def has(self, predicate: 'Predicate') -> PredicateType:
        """
        Check if any element of the current value (which must be an iterable)
        returns True when applied to predicate

        :param predicate: predicate that has to return True for at least one of this predicate's
            values
        """
        def op(v):
            for e in self.operation(v):
                if predicate(e):
                    return True
            return False
        return Predicate(op)

    inside = make_operation_two_args(operator.contains,
                                     'Return a predicate checking if x is inside value')

    instanceof = make_operation_two_args(isinstance,
                                         'Return a predicate checking whether this value '
                                         'is an instance of instance')

    length = make_operation_single_arg(len, 'Return a predicate returning length of it''s argument')

    __contains__ = make_operation_two_args(operator.contains)
    __getattr__ = make_operation_two_args(getattr)
    __getitem__ = make_operation_two_args(operator.getitem)
    __eq__ = make_operation_two_args(operator.eq)
    __ne__ = make_operation_two_args(operator.ne)
    __lt__ = make_operation_two_args(operator.lt)
    __gt__ = make_operation_two_args(operator.gt)
    __le__ = make_operation_two_args(operator.le)
    __ge__ = make_operation_two_args(operator.ge)
    __add__ = make_operation_two_args(operator.add)
    __sub__ = make_operation_two_args(operator.sub)
    __mul__ = make_operation_two_args(operator.mul)
    __matmul__ = make_operation_two_args(operator.matmul)
    __and__ = make_operation_two_args(operator.and_)
    __or__ = make_operation_two_args(operator.or_)
    __rshift__ = make_operation_two_args(operator.rshift)
    __lshift__ = make_operation_two_args(operator.lshift)
    __xor__ = make_operation_two_args(operator.xor)
    __truediv__ = make_operation_two_args(operator.__truediv__)
    __floordiv__ = make_operation_two_args(operator.floordiv)
    __mod__ = make_operation_two_args(operator.mod)
    __pow__ = make_operation_two_args(operator.pow)
    __radd__ = make_operation_two_args(operator.add, swap_order=True)
    __rsub__ = make_operation_two_args(operator.sub, swap_order=True)
    __rmul__ = make_operation_two_args(operator.mul, swap_order=True)
    __rmatmul__ = make_operation_two_args(operator.matmul, swap_order=True)
    __rand__ = make_operation_two_args(operator.and_, swap_order=True)
    __ror__ = make_operation_two_args(operator.or_, swap_order=True)
    __rrshift__ = make_operation_two_args(operator.rshift, swap_order=True)
    __rlshift__ = make_operation_two_args(operator.lshift, swap_order=True)
    __rxor__ = make_operation_two_args(operator.xor, swap_order=True)
    __rtruediv__ = make_operation_two_args(operator.__truediv__, swap_order=True)
    __rfloordiv__ = make_operation_two_args(operator.floordiv, swap_order=True)
    __rmod__ = make_operation_two_args(operator.mod, swap_order=True)
    __rpow__ = make_operation_two_args(operator.pow, swap_order=True)
    __neg__ = make_operation_single_arg(operator.neg)
    __invert__ = make_operation_single_arg(operator.invert)
    __abs__ = make_operation_single_arg(abs)


x = Predicate()



import typing as tp
import warnings

from satella.exceptions import PreconditionError

T = tp.TypeVar('T')
U = tp.TypeVar('U')
Expression = tp.NewType('Expression', str)
ExcType = tp.Type[Exception]


# noinspection PyPep8Naming
def _TRUE(x):
    return True


# taken from https://stackoverflow.com/questions/1288498/using-the-same-decorator-with-arguments-wi\
# th-functions-and-methods
def chain_functions(fun_first: tp.Callable[..., tp.Union[tp.Tuple[tp.Tuple, tp.Dict],
                                                         tp.Dict, tp.Tuple]]) -> tp.Callable:
    """
    A decorator to chain function calls.
    This function is expected to return:

     * a 2-tuple [tp.Tuple, tp.Dict] - args and kwargs for the next function
     * tp.Dict - only kwargs will be passed
     * any other tuple - only args will be passed
     * any other type - will be passed as a first argument

    of arguments to pass to wrapped function. So this

    >>> def test3(...):
    >>>     ...
    >>> def test2(...):
    >>>     ...
    >>> def test(...):
    >>>     args, kwargs = test2(...)
    >>>     return test(3)
    >>> v = test(a, b, c)

    Is equivalent to this:
    >>> @chain_functions
    >>> def test2(...):
    >>>     ...
    >>> @test2
    >>> def test3(...):
    >>>     ...
    >>> v = test3(a, b, c)
    """

    def outer(fun_next):
        @wraps(fun_next)
        def inner(*args, **kwargs):
            ret = fun_first(*args, **kwargs)
            if isinstance(ret, dict):
                args = ()
                kwargs = ret
            elif isinstance(ret, tuple) and len(ret) == 2:
                args, kwargs = ret
            elif isinstance(ret, tuple):
                args = ret
                kwargs = {}
            else:
                args = ret,
                kwargs = {}
            return fun_next(*args, **kwargs)

        return inner

    return outer


def memoize(fun):
    """
    A thread safe memoizer based on function's ONLY positional arguments.

    Note that this will make your function execute it at most one thread, the
    remaining ones will have to wait.
    """
    from satella.coding.concurrent.monitor import MonitorDict, Monitor

    fun.memoizer = MonitorDict()

    @wraps(fun)
    def inner(*args, **kwargs):
        with Monitor.acquire(fun.memoizer):
            if args in fun.memoizer:
                return fun.memoizer[args]
            else:
                v = fun(*args, **kwargs)
                fun.memoizer[args] = v
                return v
    return inner


def wraps(cls_to_wrap: tp.Type) -> tp.Callable[[tp.Type], tp.Type]:
    """
    A functools.wraps() but for classes.

    As a matter of fact, this can replace functools.wraps() entirely.
    This replaces __doc__, __name__, __module__ and __annotations__.
    It also sets a correct __wrapped__.

    :param cls_to_wrap: class to wrap
    """

    def outer(cls: tp.Type) -> tp.Type:
        if hasattr(cls_to_wrap, '__doc__'):
            cls.__doc__ = cls_to_wrap.__doc__
        if hasattr(cls_to_wrap, '__name__'):
            cls.__name__ = cls_to_wrap.__name__
        if hasattr(cls_to_wrap, '__module__'):
            cls.__module__ = cls_to_wrap.__module__
        if hasattr(cls_to_wrap, '__annotations__'):
            cls.__annotations__ = cls_to_wrap.__annotations__
        return cls

    return outer


def has_keys(keys: tp.List[str]):
    """
    A decorator for asserting that a dictionary has given keys. Will raise PreconditionError if
    it doesn't.

    This outputs a callable that accepts a dict and returns True if it has all the keys
    necessary.

    Returns True if the dict has all necessary keys.

    This is meant to be used in conjunction with @precondition

    :param keys: list of keys to expect
    """
    warnings.warn('This is deprecated and will be removed in Satella 3.0. '
                  'Use satella.coding.predicates.has_keys instead', DeprecationWarning)

    def inner(dictionary: dict) -> bool:
        for key in keys:
            if key not in dictionary:
                raise PreconditionError('Key %s not found in dict' % (key,))
        return True

    return inner


def short_none(clb: tp.Union[Expression, tp.Callable[[T], U]]) -> tp.Callable[
        [tp.Optional[T]], tp.Optional[U]]:
    """
    Accept a callable. Return a callable that executes it only if passed a no-None arg,
    and returns its result. If passed a None, return a None

    callable can also be a string, in this case it will be appended to lambda x: and eval'd

    :param clb: callable/1->1
    :return: a modified callable
    """
    if isinstance(clb, str):
        q = dict(globals())
        exec('_callable = lambda x: ' + clb, q)
        clb = q['_callable']

    @wraps(clb)
    def inner(arg: tp.Optional[T]) -> tp.Optional[U]:
        if arg is None:
            return None
        else:
            return clb(arg)

    return inner



import collections
import typing as tp

from satella.coding.recast_exceptions import silence_excs

T = tp.TypeVar('T')
KeyArg = tp.Tuple[tp.Union[int, slice], tp.Union[int, slice]]


class SparseMatrix(tp.Generic[T]):
    """
    A matrix of infinite size, that supports assignments.

    Set elements like this:

    >>> sm[1, 2] = 5
    >>> sm[:,1] = [5]
    >>> sm[1,:] = [5]
    >>> sm[:,:] = [[5]]

    where the first argument is the number of the column, counted from 0,
    and the second one is the number of the row, also counted from 0

    Note that custom slicing (ie. slices which are not :) will not be supported.
    Negative indices are supported.

    Undefined elements are considered to be of value None.

    Iterating over this matrix will yield it's consecutive rows.
    """
    def __init__(self):
        self.rows_dict = collections.defaultdict(lambda: collections.defaultdict(lambda: None))
        self.known_column_count = {}        # tp.Dict[int, int] column_no => amount
        self.no_cols = 0
        self.no_rows = 0

    def append_row(self, y: tp.Iterable[T]) -> None:
        """
        Append a row to the bottom of the matrix

        :param y: iterable with consequent columns
        """
        next_row = self.no_rows
        for col_no, z in enumerate(y):
            self[col_no, next_row] = z

    def clear(self) -> None:
        """
        Clear the contents of the sparse matrix
        """
        self.rows_dict = collections.defaultdict(lambda: collections.defaultdict(lambda: None))
        self.known_column_count = {}        # tp.Dict[int, int] column_no => amount
        self.no_cols = 0
        self.no_rows = 0

    def __eq__(self, other: 'SparseMatrix') -> bool:
        return self.rows_dict == other.rows_dict

    def __bool__(self) -> bool:
        return self.no_rows == 0

    def _sanitize_coordinate(self, coord: tp.Union[int, slice],
                             max_len: int) -> tp.Union[int, slice]:
        if isinstance(coord, int):
            if coord < 0:
                coord += max_len
        return coord

    def _sanitize_key(self, key: KeyArg) -> KeyArg:
        col, row = key

        if isinstance(col, slice):
            if not (col.start is None and col.stop is None and col.step is None):
                raise IndexError('Custom slicing is not supported!')
            col = Ellipsis
        elif isinstance(col, int):
            col = self._sanitize_coordinate(col, self.no_cols)

        if isinstance(row, slice):
            if not (row.start is None and row.stop is None and row.step is None):
                raise IndexError('Custom slicing is not supported!')
            row = Ellipsis
        elif isinstance(row, int):
            row = self._sanitize_coordinate(row, self.no_rows)

        return col, row

    @classmethod
    def from_iterable(cls, y: tp.Iterable[tp.Iterable[T]]):
        """
        Construct a sparse matrix given a row-first iterable. That iterable must
        return another iterable, that will yield values for given column.

        :param y: an iterable describing the sparse matrix
        :return: a sparse matrix object
        """
        sm = SparseMatrix()
        for row_no, cols in enumerate(y):
            for col_no, value in enumerate(cols):
                sm[col_no, row_no] = value
        return sm

    @property
    def columns(self) -> int:
        """Return the amount of columns"""
        return self.no_cols

    @property
    def rows(self) -> int:
        """Return the amount of rows"""
        return self.no_rows

    def __iter__(self) -> tp.Iterator[tp.List]:
        return (self.get_row(i) for i in range(self.no_rows))

    def __len__(self) -> int:
        return self.no_rows

    def get_row(self, row_no: int) -> tp.List[T]:
        """
        Return a single row of provided number.

        The returned array has the same length as .columns

        :param row_no: row number, numbered from 0
        """
        if row_no not in self.rows_dict:     # check so as to avoid adding new entries
            return [None]*self.no_cols
        cols = self.rows_dict[row_no]
        output = []
        for i in range(self.no_cols):
            if i in cols:
                output.append(cols[i])
            else:
                output.append(None)
        return output

    def _increment_column_count(self, col_no: int) -> None:
        if col_no not in self.known_column_count:
            self.known_column_count[col_no] = 1
        else:
            self.known_column_count[col_no] += 1

    def _decrement_column_count(self, col_no: int) -> None:
        if self.known_column_count[col_no] == 1:
            del self.known_column_count[col_no]
        else:
            self.known_column_count[col_no] -= 1

    def delete_row(self, row_no: int) -> None:
        """
        Delete a row with specified number

        :param row_no: number of the row to delete
        """
        cols = list(self.rows_dict[row_no].keys())  # Copy it here
        for col_no in cols:
            del self[col_no, row_no]

    def __setitem__(self, key: KeyArg,
                    value: tp.Union[tp.Iterable[T], T]) -> None:
        """
        Use this to update either a single piece of the matrix, or entire row.

        Use like:

        >>> sm[1, 2] = 5
        >>> sm[...,1] = [5]
        >>> sm[1,...] = [5]
        >>> sm[...,...] = [[5]]
        >>> sm[:,1] = [5]
        >>> sm[1,:] = [5]
        >>> sm[:,:] = [[5]]

        Note that setting an element to None is the same as deleting it.
        """
        if value is None:
            del self[key]
            return

        col, row = self._sanitize_key(key)

        if col is Ellipsis and row is Ellipsis:
            sm = SparseMatrix.from_iterable(value)
            self.rows_dict = sm.rows_dict
            self.no_rows = sm.no_rows
            self.no_cols = sm.no_cols
            self.known_column_count = sm.known_column_count
        elif col is Ellipsis:
            for col_no, val in enumerate(value):
                self[col_no, row] = val
        elif row is Ellipsis:
            for row_no, val in enumerate(value):
                self[col, row_no] = val
        else:
            if col >= self.no_cols:
                self.no_cols = col + 1
            if row >= self.no_rows:
                self.no_rows = row + 1

            if row not in self.rows_dict:
                self._increment_column_count(col)
            elif col not in self.rows_dict[row]:
                self._increment_column_count(col)
            self.rows_dict[row][col] = value

    def __getitem__(self,
                    item: KeyArg) -> tp.Union[tp.List[T], tp.List[tp.List[T]], T]:
        col, row = self._sanitize_key(item)

        if col is Ellipsis and row is Ellipsis:
            return list(self)
        elif col is Ellipsis:
            return [self[col_no, row] for col_no in range(self.no_cols)]
        elif row is Ellipsis:
            return [self[col, row_no] for row_no in range(self.no_rows)]
        else:
            if row not in self.rows_dict:    # check so as to avoid adding new entries
                return None
            if col not in self.rows_dict[row]:
                return None
            return self.rows_dict[row][col]

    @silence_excs(TypeError, returns=0)
    def _calculate_column_count(self) -> int:
        return max(self.known_column_count) + 1

    @silence_excs(TypeError, returns=0)
    def _calculate_row_count(self) -> int:
        return max(self.rows_dict) + 1

    @silence_excs(KeyError)
    def __delitem__(self, key: KeyArg) -> None:
        col, row = self._sanitize_key(key)

        if row is Ellipsis and col is Ellipsis:
            self.clear()
        elif col is Ellipsis:
            for col_no in range(self.no_cols):
                del self[col_no, row]
        elif row is Ellipsis:
            for row_no in range(self.no_rows):
                del self[col, row_no]
        else:
            # Check if the element is there
            if row not in self.rows_dict:
                return

            if col not in self.rows_dict[row]:
                return

            del self.rows_dict[row][col]

            if not self.rows_dict[row]:     # Have we got an empty row now?
                del self.rows_dict[row]

            self._decrement_column_count(col)

            self.no_cols = self._calculate_column_count()
            self.no_rows = self._calculate_row_count()



from concurrent.futures import Future

from .common import ResponseFuture


def wrap_future(future: ResponseFuture) -> Future:
    """
    Convert a Cassandra's future to a normal Python future.

    :param future: cassandra future to wrap
    :return: a standard Python future
    :raises TypeError: when a normal Python future is passed as future
    """
    if isinstance(future, Future):
        raise TypeError('Tried to wrap an existing Future!')

    fut = Future()
    fut.set_running_or_notify_cancel()
    future.add_callback(lambda result: fut.set_result(result))
    future.add_errback(lambda exception: fut.set_exception(exception))
    return fut




import enum
import typing as tp
from abc import ABCMeta, abstractmethod


class HashableIntEnum(enum.IntEnum):
    """
    An enum.IntEnum that implements hashability, stemming from it's values, as well
    as hashability
    """

    def __hash__(self) -> int:
        return hash(self.value)


class ComparableIntEnum(HashableIntEnum):
    """
    An enum.IntEnum that implements comparision, stemming from it's values, as well
    as hashability
    """

    def __hash__(self) -> int:
        return hash(self.value)

    def __lt__(self, other: 'ComparableIntEnum') -> bool:
        return self.value < other.value

    def __le__(self, other: 'ComparableIntEnum') -> bool:
        return self.value <= other.value

    def __gt__(self, other: 'ComparableIntEnum') -> bool:
        return self.value > other.value

    def __ge__(self, other: 'ComparableIntEnum') -> bool:
        return self.value >= other.value

    def __eq__(self, other: 'ComparableIntEnum') -> bool:
        return self.value == other.value


class StrEqHashableMixin(metaclass=ABCMeta):
    """
    A mix-in that outfits your class with an __eq__ and __hash__ operator that take
    their values from __str__ representation of your class.
    """
    __slots__ = ()

    def __eq__(self, other) -> bool:
        return str(self) == str(other)

    def __hash__(self) -> int:
        return hash(str(self))

    @abstractmethod
    def __str__(self) -> str:
        pass


class ComparableAndHashableBy(metaclass=ABCMeta):
    """
    A mix-in. Provides comparision (lt, gt, ge, le, eq) and hashing by a field of this class.

    Example:

    >>> class Vector(ComparableAndHashableBy):
    >>>     _COMPARABLE_BY = 'length'
    >>>     @property
    >>>     def length(self):
    >>>         ...
    >>>
    >>> assert Vector() > Vector()
    """
    __slots__ = ()

    @property
    @abstractmethod
    def _COMPARABLE_BY(self) -> str:
        """
        Return the sequence of names of properties and attributes
        that will be used for __eq__ and __hash__
        """
        return ''

    def __hash__(self):
        return hash(getattr(self, self._COMPARABLE_BY))

    def __eq__(self, other: 'ComparableAndHashableBy') -> bool:
        return getattr(self, self._COMPARABLE_BY) == getattr(other, other._COMPARABLE_BY)

    def __lt__(self, other: 'ComparableAndHashableBy') -> bool:
        return getattr(self, self._COMPARABLE_BY) < getattr(other, other._COMPARABLE_BY)

    def __le__(self, other: 'ComparableAndHashableBy') -> bool:
        return getattr(self, self._COMPARABLE_BY) <= getattr(other, other._COMPARABLE_BY)

    def __gt__(self, other: 'ComparableAndHashableBy') -> bool:
        return getattr(self, self._COMPARABLE_BY) > getattr(other, other._COMPARABLE_BY)

    def __ge__(self, other: 'ComparableAndHashableBy') -> bool:
        return getattr(self, self._COMPARABLE_BY) >= getattr(other, other._COMPARABLE_BY)


class ComparableAndHashableByInt(metaclass=ABCMeta):
    """
    A mix-in. Provides comparision (lt, gt, ge, le, eq) and hashing by __int__ of this class.
    """

    __slots__ = ()

    @abstractmethod
    def __int__(self) -> int:
        ...

    def __hash__(self):
        return hash(int(self))

    def __eq__(self, other: 'ComparableAndHashableByInt') -> bool:
        return int(self) == int(other)

    def __lt__(self, other: 'ComparableAndHashableByInt') -> bool:
        return int(self) < int(other)

    def __le__(self, other: 'ComparableAndHashableByInt') -> bool:
        return int(self) <= int(other)

    def __gt__(self, other: 'ComparableAndHashableByInt') -> bool:
        return int(self) > int(other)

    def __ge__(self, other: 'ComparableAndHashableByInt') -> bool:
        return int(self) >= int(other)


class OmniHashableMixin(metaclass=ABCMeta):
    """
    A mix-in. Provides hashing and equal comparison for your own class using specified fields.

    Example of use:

    >>> class Point2D(OmniHashableMixin):
    >>>    _HASH_FIELDS_TO_USE = ('x', 'y')
    >>>    def __init__(self, x, y):
    >>>        ...

    and now class Point2D has defined __hash__ and __eq__ by these fields.
    Do everything in your power to make specified fields immutable, as mutating them will result
    in a different hash.

    _HASH_FIELDS_TO_USE can be also a single string, in this case a single field called by this
    name will be taken.

    Note that if you're explicitly providing __eq__ in your child class, you will be required to
    insert:

    >>>     __hash__ = OmniHashableMixin.__hash__

    for this to work in your class
    """
    __slots__ = ()

    @property
    @abstractmethod
    def _HASH_FIELDS_TO_USE(self) -> tp.Union[str, tp.Sequence[str]]:
        """
        Return the sequence of names of properties and attributes
        that will be used for __eq__ and __hash__
        """
        return ()

    def __hash__(self) -> int:
        xor = 0
        cmpr_by = self._HASH_FIELDS_TO_USE
        if isinstance(cmpr_by, str):
            return hash(getattr(self, cmpr_by))

        for field_name in cmpr_by:
            xor ^= hash(getattr(self, field_name))
        return xor

    def __eq__(self, other) -> bool:
        """
        Note that this will only compare _HASH_FIELDS_TO_USE
        """
        if isinstance(other, OmniHashableMixin):
            cmpr_by = self._HASH_FIELDS_TO_USE

            if isinstance(cmpr_by, str):
                return getattr(self, cmpr_by) == getattr(other, cmpr_by)

            for field_name in self._HASH_FIELDS_TO_USE:
                if getattr(self, field_name) != getattr(other, field_name):
                    return False
            return True
        else:
            return super().__eq__(other)

    def __ne__(self, other) -> bool:
        if isinstance(other, OmniHashableMixin):
            cmpr_by = self._HASH_FIELDS_TO_USE

            if isinstance(cmpr_by, str):
                return getattr(self, cmpr_by) != getattr(other, cmpr_by)

            for field_name in cmpr_by:
                if getattr(self, field_name) != getattr(other, field_name):
                    return True
            return False
        else:
            return super().__ne__(other)


class ReprableMixin:
    """
    A sane __repr__ default.

    This takes the values for the __repr__ from repr'ing list of fields defined as
    class property _REPR_FIELDS.

    Set an optional class property of _REPR_FULL_CLASSNAME for __repr__ to output the repr alongside the module name.

    Example:

    >>> class Test(ReprableMixin):
    >>>     _REPR_FIELDS = ('v', )
    >>>     def __init__(self, v, **kwargs):
    >>>         self.v = v
    >>>
    >>> assert repr(Test(2)) == "Test(2)"
    >>> assert repr(Test('2') == "Test('2')")
    """
    __slots__ = ()
    _REPR_FIELDS = ()

    def __repr__(self):
        fragments = []
        if getattr(self, '_REPR_FULL_CLASSNAME', False):
            fragments = ['%s%s' % ((self.__class__.__module__ + '.')
                                   if self.__class__.__module__ != 'builtins' else '',
                                   self.__class__.__qualname__)]
        if not fragments:
            fragments = [self.__class__.__name__]
        fragments.append('(')
        arguments = []
        for field_name in self._REPR_FIELDS:
            arguments.append(repr(getattr(self, field_name)))
        fragments.append(', '.join(arguments))
        fragments.append(')')
        return ''.join(fragments)



import codecs
import io
import os
import re
import typing as tp
import shutil

__all__ = ['read_re_sub_and_write', 'find_files', 'split', 'read_in_file', 'write_to_file',
           'write_out_file_if_different', 'make_noncolliding_name', 'try_unlink',
           'DevNullFilelikeObject']


from satella.coding import silence_excs

SEPARATORS = {'\\', '/'}
SEPARATORS.add(os.path.sep)


class DevNullFilelikeObject:
    """
    A /dev/null filelike object. For multiple uses.
    """
    __slots__ = ('is_closed', )

    def __init__(self):
        self.is_closed = False

    def read(self, byte_count: tp.Optional[int] = None):
        """
        :raises ValueError: this object has been closed
        :raises io.UnsupportedOperation: since reading from this is forbidden
        """
        if self.is_closed:
            raise ValueError('Reading from closed /dev/null!')
        raise io.UnsupportedOperation('read')

    def write(self, x: tp.Union[str, bytes]) -> int:
        """
        Discard any amount of bytes

        :raises ValueError: this object has been closed
        :return: length of written content
        """
        if self.is_closed:
            raise ValueError('Writing to closed /dev/null!')
        return len(x)

    def flush(self) -> None:
        """
        :raises ValueError: when this object has been closed
        """
        if self.is_closed:
            raise ValueError('flush of closed file')

    def close(self) -> None:
        """
        Close this stream. Further write()s and flush()es will raise a ValueError.
        No-op if invoked multiple times
        """
        self.is_closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def _has_separator(path: str) -> bool:
    # handle Windows case
    if len(path) == 3:
        if path.endswith(':/') or path.endswith(':\\'):
            return False
    return any(map(lambda x: x in path, SEPARATORS))


def make_noncolliding_name(path: str,
                           exists_checker: tp.Callable[[str], bool] = os.path.exists) -> str:
    """
    Try to make a noncolliding name in such a way that .1, .2, .3, and so on will be appended
    to the file name right before the extension (yielding test.1.txt) or at the end of the file
    name if the extension isn't present

    :param path: path of the file that has not to exist
    :param exists_checker: a callable to check with if the file exists
    :return: name mutated in such a way that exists_checker returned False on it
    """
    path, filename = os.path.split(path)
    if '.' in filename:
        *filename, extension = filename.split('.')
        filename = '.'.join(filename)
        extension = '.'+extension
    else:
        extension = ''
    addition = ''
    addition_counter = 0
    while exists_checker(os.path.join(path, filename+addition+extension)):
        addition_counter += 1
        addition = '.' + str(addition_counter)
    return os.path.join(path, filename+addition+extension)


def split(path: str) -> tp.List[str]:
    """
    An exact reverse of os.path.join

    Is is true that

    >>> os.path.join(split(a)) == a
    """
    data = list(os.path.split(path))
    while _has_separator(data[0]):
        data = list(os.path.split(data[0])) + data[1:]
    return data


def write_to_file(path: str, data: tp.Union[bytes, str],
                  encoding: tp.Optional[str] = None) -> None:
    """
    Write provided content as a file, applying given encoding (or data is bytes, if none given)

    :param path: Path to put the file under
    :param data: Data to write. Must be bytes if no encoding is given, str otherwise
    :param encoding: Encoding. Default is None, which means no encoding (bytes will be written)
    """
    if encoding is None:
        file = open(path, 'wb')
    else:
        file = codecs.open(path, 'wb', encoding)

    try:
        file.write(data)
    finally:
        file.close()


def read_in_file(path: str, encoding: tp.Optional[str] = None,
                 default: tp.Optional[tp.Union[bytes, str]] = None) -> tp.Union[bytes, str]:
    """
    Opens a file for reading, reads it in, converts to given encoding (or returns as bytes
    if not given), and closes it.

    :param path: path of file to read
    :param encoding: optional encoding. If default (None given), this will be returned as bytes
    :param default: value to return when the file does not exist. Default (None) will raise a
        FileNotFoundError
    :return: file content, either decoded as a str, or not as bytes
    """
    if os.path.isdir(path):
        if default:
            return default
        raise FileNotFoundError('%s found and is a directory' % (path, ))

    try:
        if encoding is None:
            file = open(path, 'rb')
        else:
            file = codecs.open(path, 'rb', encoding)
    except FileNotFoundError:
        if default:
            return default
        raise

    try:
        return file.read()
    finally:
        file.close()


def read_re_sub_and_write(path: str, pattern: tp.Union[re.compile, str],
                          repl: tp.Union[tp.Callable[[tp.Any], str]]) -> None:
    """
    Read a text file, treat with re.sub and write the contents

    :param path: path of file to treat
    :param pattern: regexp compiled pattern or a string, a pattern to match the file contents
    :param repl: string or a callable(re.Match)->str to replace the contents
    """
    with open(path, 'r') as f_in:
        data = f_in.read()

    if isinstance(pattern, str):
        data = re.sub(pattern, repl, data)
    else:
        data = pattern.sub(repl, data)

    with open(path, 'w') as f_out:
        f_out.write(data)


@silence_excs(OSError, returns=False)
def try_unlink(path: str) -> bool:
    """
    A syntactic sugar for:

    >>> try:
    >>>     os.unlink(path)
    >>>     return True
    >>> except FileNotFoundError:
    >>>     return False

    Note that if path is a directory, rmtree from shlex will be called on it, and
    any OSErrors will report the deletion as False

    :param path: path of file to delete
    :return: whether the deletion happened
    """
    if os.path.isdir(path):
        shutil.rmtree(path)
    else:
        os.unlink(path)
    return True


def _cond_join(prefix: tp.Optional[str], filename: str) -> str:
    """or a conditional os.path.join"""
    if prefix is None:
        return filename
    else:
        return os.path.join(prefix, filename)


def find_files(path: str, wildcard: str = r'(.*)',
               prefix_with: tp.Optional[str] = None,
               scan_subdirectories: bool = True,
               apply_wildcard_to_entire_path: bool = False,
               prefix_with_path: bool = True) -> tp.Iterator[str]:
    """
    Look at given path's files and all subdirectories and return an iterator of
    file names (paths included) that conform to given wildcard.

    Note that wildcard is only applied to the file name if apply_wildcard_to_entire_path
    is False, else the wildcard is applied to entire path (including the application of
    prefix_with!).

    Files will be additionally prefixed with path, but only if prefix_with_path is True

    :param path: path to look into.
    :param wildcard: a regular expression to match
    :param prefix_with: an optional path component to prefix before the filename with os.path.join
    :param scan_subdirectories: whether to scan subdirectories
    :param apply_wildcard_to_entire_path: whether to take the entire relative path into account
        when checking wildcard
    :param prefix_with_path: whether to add path to the resulting path
    :return: paths with the files. They will be relative paths, relative to path
    """
    if prefix_with_path:
        prefix_with = _cond_join(prefix_with, path)

    for filename in os.listdir(path):
        if scan_subdirectories and os.path.isdir(os.path.join(path, filename)):
            new_prefix = _cond_join(prefix_with, filename)
            yield from find_files(os.path.join(path, filename), wildcard,
                                  prefix_with=new_prefix,
                                  prefix_with_path=False)
        else:
            if apply_wildcard_to_entire_path:
                fn_path = _cond_join(prefix_with, filename)
            else:
                fn_path = filename
            if re.match(wildcard, fn_path):
                yield _cond_join(prefix_with, filename)


def write_out_file_if_different(path: str, data: tp.Union[bytes, str],
                                encoding: tp.Optional[str] = None) -> bool:
    """
    Syntactic sugar for

    >>> try:
    >>>     if read_in_file(path, encoding) != data:
    >>>         write_to_file(path, data, encoding)
    >>>         return True
    >>>     else:
    >>>         return False
    >>> except OSError:
    >>>     write_to_file(path, data, encoding)
    >>>     return True

    :param path: Path to put the file under
    :param data: Data to write. Must be bytes if no encoding is given, str otherwise
    :param encoding: Encoding. Default is None, which means no encoding (bytes will be written)
    :return: if write has happened
    """
    try:
        if read_in_file(path, encoding) != data:
            write_to_file(path, data, encoding)
            return True
        else:
            return False
    except FileNotFoundError:
        write_to_file(path, data, encoding)
        return True
